use std::{collections::HashMap, sync::LazyLock};

use bollard::{
    Docker,
    image::CreateImageOptions,
    query_parameters::{
        CreateContainerOptionsBuilder, CreateImageOptionsBuilder, ListContainersOptionsBuilder,
        StartContainerOptions,
    },
    secret::ContainerCreateBody,
};
use futures_util::TryStreamExt;

use crate::state::ClientState;

#[derive(Clone)]
pub struct ContainerRuntime {
    client: Docker,
}

impl ContainerRuntime {
    #[tracing::instrument(skip(self), level = "trace")]
    pub async fn ensure_running(&self, name: &str, image: &str) -> anyhow::Result<()> {
        tracing::debug!("ensuring that image is running");

        let containers = self
            .client
            .list_containers(Some(
                ListContainersOptionsBuilder::default()
                    .all(true)
                    .filters(&HashMap::from([(
                        "name".to_string(),
                        vec![name.to_string()],
                    )]))
                    .build(),
            ))
            .await?;
        if !containers.is_empty() {
            // Reconcile difference
            return Ok(());
        }

        let _ = self
            .client
            .create_image(
                Some(CreateImageOptionsBuilder::new().from_image(image).build()),
                None,
                None,
            )
            .try_collect::<Vec<_>>()
            .await?;

        self.client
            .create_container(
                Some(CreateContainerOptionsBuilder::new().name(name).build()),
                ContainerCreateBody {
                    image: Some(image.into()),
                    ..Default::default()
                },
            )
            .await?;

        self.client
            .start_container(name, None::<StartContainerOptions>)
            .await?;

        Ok(())
    }
}

pub trait ContainerRuntimeState {
    fn container_runtime(&self) -> ContainerRuntime;
}

impl ContainerRuntimeState for ClientState {
    fn container_runtime(&self) -> ContainerRuntime {
        static CLIENT: LazyLock<bollard::Docker> = LazyLock::new(|| {
            Docker::connect_with_defaults().expect("to be able to connect to a docker daemon")
        });

        ContainerRuntime {
            client: CLIENT.clone(),
        }
    }
}
